/*!	\file		CicadaBest.cpp
	\author		Garth Santor
	\date		2017-01-05, 2019-02-12
	\version	1.0.0

	Cycada Cycles solution in C++.
*/

#include <algorithm>
#include <iostream>
#include <vector>
#include <locale>
using namespace std;


// TYPES/Limits
using count_t = std::uintmax_t;
count_t constexpr MAX_CICADA_CYCLES = 20;


/*!	Cicada trial result.
*/
struct CicadaTrialResult {
	count_t cycleYears;
	count_t numberPredatorYears;
};
bool operator < (CicadaTrialResult const& lhs, CicadaTrialResult const& rhs) {
	return lhs.numberPredatorYears < rhs.numberPredatorYears;
}


/*!	Count the number of years that the cicada will coincide with predators.

	\param kCicadaCycle [in] the number of years in the cicada's life cycle.
	\param kTrialYears [in] the number of years to test against the cicada's cycle.
	\return the number of predators that align with the cicada's life cycle.
*/
CicadaTrialResult num_predators(count_t const kCicadaCycle, count_t const kTrialYears) {
	count_t predatorCount = 0;
	for (count_t year = 1; year <= kTrialYears; ++year) {
		if (year%kCicadaCycle == 0) {
			predatorCount += ((year % 2) == 0);
			predatorCount += ((year % 3) == 0);
			predatorCount += ((year % 4) == 0);
			predatorCount += ((year % 6) == 0);
		}
	}
	return CicadaTrialResult{ kCicadaCycle, predatorCount };
}



/*!	Process entry point - run the cicada trials.
*/
int main() {
	// Query the user for the number of trials to perform.
	cout << "Enter the number of years to test: ";
	uintmax_t trialYears;
	cin >> trialYears;
	if (!cin) {
		cerr << "Error: failed to get a valid number.\n";
		return EXIT_FAILURE;
	}
	if (trialYears <= 0) {
		cerr << "Error: the number of trial years must be a natural number.\n";
		return EXIT_FAILURE;
	}


	// Make an array of results.
	vector<CicadaTrialResult> results;
	for (int cicadaCycle = 1; cicadaCycle <= MAX_CICADA_CYCLES; ++cicadaCycle)
		results.push_back(num_predators(cicadaCycle, trialYears));


	// Output a table of trial results.
	cout.imbue(locale(""));
	auto print = [](auto& result) { cout << result.cycleYears << "\t\t" << result.numberPredatorYears << '\n'; };
	cout << "cicada cycle\t#predators\n";
	for (auto result : results)
		print(result);


	// Output a table of the best 4 results
	sort(begin(results), end(results));
	cout << "\n\n\nBest Cycles\n\ncicada cycle\t#predators\n";
	std::for_each(begin(results), begin(results) + 4, print);
}